import logging
from typing import Optional, Tuple

import requests
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

logger = logging.getLogger(__name__)


class ChromeWebDriver:
    """
    ChromeWebDriver class to fetch the HTTP response for a given URL

    Implementation of the Selenium Chrome WebDriver
    with HTTP Response data included via the ChromeDriver performance logging capability
    https://sites.google.com/a/chromium.org/chromedriver/logging/performance-log
    The ChromeWebDriver response attribute(s) contain a dict with information about the response
    {
        "connectionId": [Integer],
        "connectionReused": [Boolean],
        "encodedDataLength": [Integer],
        "fromDiskCache": [Boolean],
        "fromServiceWorker": [Boolean],
        "headers": [dict], # HTTP Headers as a dict
        "headersText": [String], # HTTP Headers as text
        "mimeType": [String],
        "protocol": [String],
        "remoteIPAddress": [String],
        "remotePort": [Integer],
        "requestHeaders": [dict],
        "requestHeadersText": [String],
        "securityDetails": [dict], # TLS/SSL related information
        "securityState": [String],
        "status": [Integer], # HTTP Status Code of the Response
        "statusText": [String],
        "timing": [dict],
        "url": [String]
    }
    """

    autoclose: bool = True
    options: Options = None
    status_code: int = 0
    text: str = ""
    title: str = ""
    url: str
    wait_time: int = 10

    _driver: Optional[webdriver.Chrome] = None

    def __init__(
        self,
        url: Optional[str] = None,
        headless: bool = True,
        no_sandbox: bool = True,
        disable_dev_shm_usage: bool = True,
        autoclose: bool = True,
        wait_time: int = 10,
    ):
        """
        Initializes a ChromeWebDriver instance.

        Args:
            url (Optional[str]): URL to fetch. Defaults to None.
            headless (bool): Whether to run the ChromeWebDriver in headless mode. Defaults to True.
            no_sandbox (bool): Whether to disable the Chrome sandbox. Defaults to True.
            disable_dev_shm_usage (bool): Whether to disable the /dev/shm usage. Defaults to True.
            autoclose (bool): Whether to automatically close the ChromeWebDriver instance. Defaults to True.
            wait_time (int): The maximum time to wait for the page to load, in seconds. Defaults to 1.
        """
        self.url = url  # URL to fetch
        self.options = Options()
        if headless:
            self.options.add_argument("--headless")
        if no_sandbox:
            self.options.add_argument("--no-sandbox")
        if disable_dev_shm_usage:
            self.options.add_argument("--disable-dev-shm-usage")
        self.autoclose = autoclose
        self.wait_time = wait_time

        self._driver = None
        self.text = ""
        self.title = ""
        self.status_code = 0
        if url is not None:
            self.get(url)

    @property
    def driver(self):
        """
        Property that returns the ChromeWebDriver instance.

        Returns:
            ChromeWebDriver: The ChromeWebDriver instance.
        """
        if self._driver is None or self._driver.session_id is None:
            self._driver = webdriver.Chrome(options=self.options)
        return self._driver

    def get(
        self,
        url: str,
        wait_time: Optional[int] = None,
        locator: Optional[Tuple[str, str]] = None,
    ) -> "ChromeWebDriver":
        """
        Fetches the HTTP response for the given URL using the ChromeWebDriver.

        This method sets the URL to fetch and attempts to fetch the response.
        If an error occurs during the fetch, it logs the error and closes the ChromeWebDriver instance.

        Args:
            url (str): The URL to fetch.
            wait_time (int, optional): The maximum time to wait for the page to load, in seconds. Defaults to 10.
            wait_for_element_id (str, optional): The ID of the element to wait for. Defaults to None.

        Returns:
            ChromeWebDriver: The ChromeWebDriver instance.
        """
        self.url = url  # set the URL to fetch
        wait_time = wait_time or self.wait_time
        try:
            self._get(
                url,
                wait_time,
                locator=locator,
            )
        except Exception as e:
            logger.error("Error while fetching the url: %s", url)
            logger.error(e)
            self.close()
        return self

    def _get(
        self,
        url: str,
        wait_time: int = 10,
        locator: Optional[Tuple[str, str]] = None,
    ):
        self.status_code = requests.get(url).status_code  # get the HTTP status code
        self.driver.get(url)  # get the requested URL
        self.driver.refresh()
        if wait_time > 0 and locator is not None:
            WebDriverWait(self.driver, wait_time).until(
                EC.presence_of_element_located(locator)
            )
        self.title = self.driver.title
        self.text = self.driver.page_source
        if self.autoclose:
            self.close()

    def close(self):
        """
        Closes the ChromeWebDriver instance.

        This method closes the ChromeWebDriver instance by closing the current window
        and quitting the driver, releasing all associated resources.
        """
        if self._driver is not None:
            self._driver.close()



import logging
from datetime import datetime
from typing import List, Optional, Tuple

from bs4 import BeautifulSoup

from .base import BaseFetcher, By

logger = logging.getLogger(__name__)


class BisFetcher(BaseFetcher):
    """
    Fetcher for BIS.
    """

    _config_name_: str = "bis"
    _config_group_: str = "/fetcher"
    output_dir: str = f"workspace/datasets{_config_group_}/{_config_name_}"

    base_url: str = "https://www.bis.org"
    search_url: str = (
        "https://www.bis.org/cbspeeches/index.htm?m=256&cbspeeches_page={page}"
    )
    search_keywords: List[str] = []

    link_locator: Tuple[str, str] = (
        By.CSS_SELECTOR,
        "#cbspeeches_list > div > table > tbody > tr:nth-child(10) > td:nth-child(2) > div > div.title",
    )
    link_container_name: str = "table"
    link_container_attrs: dict = {"class": "documentList"}
    link_find_all_name: str = "tr"
    link_find_all_attrs: dict = {}
    lint_article_name: str = "div"
    lint_article_attrs: dict = {"class": "title"}

    def _parse_page_links(
        self,
        page_url: str,
        print_every: int = 10,
        verbose: bool = False,
    ) -> Optional[List[dict]]:
        """Get the links from the given page."""
        links = []
        try:
            response = self.request(
                page_url,
                use_selenium=True,
                locator=self.link_locator,
            )
            # Check if page exists (status code 200) or not (status code 404)
            if response.status_code == 404:
                logger.info("Page [%s] does not exist, stopping...", page_url)
                return None
            soup = BeautifulSoup(response.text, "html.parser")

            # Find the section with class 'section-category'
            section = soup.find(
                self.link_container_name, attrs=self.link_container_attrs
            )

            # Find all articles within the section
            articles = section.find_all(
                self.link_find_all_name, attrs=self.link_find_all_attrs
            )

            for article_no, article in enumerate(articles):
                # Extract and print article information
                title_div = article.find(
                    self.lint_article_name, attrs=self.lint_article_attrs
                )
                if title_div is None:
                    logger.info("No title found for article %s", article_no)
                    continue
                title = title_div.text
                url = self.base_url + article.find("a")["href"]

                date_ = article.find("td", class_="item_date")
                item_date = date_.text.strip() if date_ else ""
                author_ = article.find("a", class_="authorlnk dashed")
                author = author_.text.strip() if author_ else ""

                if verbose and article_no % print_every == 0:
                    logger.info("Title: %s", title)
                    logger.info("URL: %s", url)
                link = {
                    "title": title,
                    "author": author,
                    "timestamp": item_date,
                    "url": url,
                }
                links.append(link)
        except Exception as e:
            logger.error("Error while fetching the page url: %s", page_url)
            logger.error(e)
        return links

    def _parse_article_text(self, url: str) -> Optional[dict]:
        """Parse the article text from the given divs."""
        try:
            if url.endswith(".pdf"):
                return {
                    "pdf_url": url,
                }

            response = self.request(url)
            soup = BeautifulSoup(response.text, "html.parser")
            pdf_div = soup.find("div", class_="pdftxt")
            pdf = pdf_div.find("a", class_="pdftitle_link")["href"]
            pdf_url = self.base_url + pdf

            return {
                "pdf_url": pdf_url,
            }

        except Exception as e:
            logger.error("Error while scraping the article url: %s", url)
            logger.error(e)
        return None



"""Base Fetcher"""
import logging
import multiprocessing as mp
import time
from functools import partial
from pathlib import Path
from typing import Callable, List, Optional, Tuple

import requests
from hyfi.composer import BaseModel
from hyfi.main import HyFI

from .chromedriver import ChromeWebDriver

logger = logging.getLogger(__name__)


class Response(BaseModel):
    text: str = ""
    status_code: int = 0


class By:
    """Set of supported locator strategies."""

    ID = "id"
    XPATH = "xpath"
    LINK_TEXT = "link text"
    PARTIAL_LINK_TEXT = "partial link text"
    NAME = "name"
    TAG_NAME = "tag name"
    CLASS_NAME = "class name"
    CSS_SELECTOR = "css selector"


class BaseFetcher(BaseModel):
    """
    Base Fetcher
    """

    _config_name_: str = "base"
    _config_group_: str = "/fetcher"

    article_filename: str = "articles.jsonl"
    base_url: str = ""
    delay_between_requests: float = 0.0
    key_field: str = "url"
    keyword_placeholder: str = "{keyword}"
    link_filename: str = "links.jsonl"
    max_num_articles: Optional[int] = 30
    max_num_pages: Optional[int] = 2
    num_workers: int = 1
    output_dir: str = f"workspace/datasets{_config_group_}/{_config_name_}"
    overwrite_existing: bool = False
    page_placeholder: str = "{page}"
    print_every: int = 10
    search_keywords: List[str] = []
    search_url: str = ""
    start_page: Optional[int] = 1
    start_urls: List[str] = []
    verbose: bool = True

    _links: List[dict] = []
    _articles: List[dict] = []
    _headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3"
    }

    def __call__(self):
        self.fetch()

    def fetch(self):
        self.fetch_links()
        self.fetch_articles()

    def request(
        self,
        url: str,
        use_selenium: bool = False,
        wait_time: int = 10,
        locator: Optional[Tuple[str, str]] = None,
        params: dict = None,
        **kwargs,
    ) -> Response:
        """Sends a GET request.

        Args:
            url (str): URL for the request
            params (dict, optional): Dictionary, list of tuples or bytes to send
                in the query string for the Request. Defaults to None.
            **kwargs: Optional arguments that `request` takes.

        Returns:
            Response object containing response text and status code
        """
        if use_selenium:
            driver = ChromeWebDriver().get(
                url,
                wait_time=wait_time,
                locator=locator,
            )
            return Response(text=driver.text, status_code=driver.status_code)
        res = requests.get(url, params=params, headers=self._headers, **kwargs)
        return Response(text=res.text, status_code=res.status_code)

    @property
    def start_urls_encoded(self):
        if self.start_urls:
            return self.start_urls
        if self.keyword_placeholder in self.search_url:
            # Generate start URLs by replacing keyword placeholder with encoded keywords
            start_urls = [
                self.search_url.replace(
                    self.keyword_placeholder, self.encode_keyword(keyword)
                )
                for keyword in self.search_keywords
            ]
        else:
            start_urls = [self.search_url]
        return start_urls

    def encode_keyword(self, keyword: str):
        return keyword.replace(" ", "+")

    @property
    def links(self):
        return self._links or self._load_links()

    @property
    def articles(self):
        return self._articles or self._load_articles()

    @property
    def link_filepath(self) -> str:
        _path = Path(self.output_dir) / self.link_filename
        _path.parent.mkdir(parents=True, exist_ok=True)
        return str(_path.absolute())

    @property
    def article_filepath(self) -> str:
        _path = Path(self.output_dir) / self.article_filename
        _path.parent.mkdir(parents=True, exist_ok=True)
        return str(_path.absolute())

    @property
    def link_filepath_tmp(self) -> str:
        _path = Path(self.output_dir) / f"{self.link_filename}.tmp"
        _path.parent.mkdir(parents=True, exist_ok=True)
        return str(_path)

    @property
    def article_filepath_tmp(self) -> str:
        _path = Path(self.output_dir) / f"{self.article_filename}.tmp"
        _path.parent.mkdir(parents=True, exist_ok=True)
        return str(_path)

    def _load_links(self) -> List[dict]:
        if Path(self.link_filepath).exists():
            self._links = HyFI.load_jsonl(self.link_filepath)
        return self._links

    def _load_articles(self) -> List[dict]:
        if Path(self.article_filepath).exists():
            self._articles = HyFI.load_jsonl(self.article_filepath)
        return self._articles

    def fetch_links(self):
        parse_page_func = partial(
            self._parse_page_links,
            print_every=self.print_every,
            verbose=self.verbose,
        )

        next_page_func = partial(
            self._next_page_func,
            page_placeholder=self.page_placeholder,
        )

        self._fetch_links(parse_page_func, next_page_func)

    def fetch_articles(self):
        parse_article_text = partial(self._parse_article_text)

        self._fetch_articles(parse_article_text)

    def _fetch_links(self, parse_page_func: Callable, next_page_func: Callable):
        num_workers = min(self.num_workers, len(self.search_keywords))
        num_workers = max(num_workers, 1)
        link_urls = [link["url"] for link in self.links]
        fetch_links_func = partial(
            crawl_links,
            parse_page_func=parse_page_func,
            next_page_func=next_page_func,
            start_page=self.start_page,
            max_num_pages=self.max_num_pages,
            link_urls=link_urls,
            link_filepath=self.link_filepath_tmp,
            delay_between_requests=self.delay_between_requests,
        )
        if links := self._fetch_links_mp(
            num_workers,
            fetch_links_func,
        ):
            self.save_links(links)
        else:
            logger.info("No more links found")

    def save_links(self, links: List[dict]):
        self._links.extend(links)
        original_len = len(self._links)
        self._links = HyFI.remove_duplicates_from_list_of_dicts(
            self._links, key=self.key_field
        )
        logger.info(
            "Removed %s duplicate links from %s links",
            original_len - len(self._links),
            original_len,
        )
        HyFI.save_jsonl(self._links, self.link_filepath)
        logger.info("Saved %s links to %s", len(self._links), self.link_filepath)

    def _fetch_links_mp(
        self,
        num_workers: int,
        batch_func: Callable,
    ) -> List[dict]:
        with mp.Pool(num_workers) as pool:
            results = pool.map(batch_func, self.start_urls_encoded)
        links = []
        for result in results:
            links.extend(result)
        return links

    def _fetch_articles(self, parse_article_func: Callable):
        num_workers = min(self.num_workers, len(self.links))
        article_urls = [article["url"] for article in self.articles]
        fetch_articles_func = partial(
            scrape_article_text,
            parse_article_func=parse_article_func,
            article_urls=article_urls,
            overwrite_existing=self.overwrite_existing,
            article_filepath=self.article_filepath_tmp,
            max_num_articles=self.max_num_articles,
            delay_between_requests=self.delay_between_requests,
            print_every=self.print_every,
            verbose=self.verbose,
        )
        if articles := self._fetch_articles_mp(num_workers, fetch_articles_func):
            self.save_articles(articles)
        else:
            logger.info("No more articles found")

    def save_articles(self, articles: List[dict]):
        self._articles.extend(articles)
        original_len = len(self._articles)
        self._articles = HyFI.remove_duplicates_from_list_of_dicts(
            self._articles, key=self.key_field
        )
        logger.info(
            "Removed %s duplicate articles from %s articles",
            original_len - len(self._articles),
            original_len,
        )
        HyFI.save_jsonl(self._articles, self.article_filepath)
        logger.info(
            "Saved %s articles to %s",
            len(self._articles),
            self.article_filepath,
        )

    def _fetch_articles_mp(
        self,
        num_workers: int,
        batch_func: Callable,
    ) -> List[dict]:
        articles = []
        if len(self.links) < 1:
            return articles
        batch_size = len(self.links) // num_workers
        batches = [
            self.links[i : i + batch_size]
            for i in range(0, len(self.links), batch_size)
        ]
        with mp.Pool(num_workers) as pool:
            results = pool.map(batch_func, batches)
        for result in results:
            articles.extend(result)
        return articles

    def _next_page_func(
        self,
        start_url: str,
        current_url: Optional[str],
        page: int,
        page_placeholder: Optional[str],
    ) -> Optional[str]:
        if page_placeholder and page_placeholder in start_url:
            # Return next page url by replacing placeholder with page number
            page_url = start_url.replace(page_placeholder, str(page))
        elif current_url is None:
            page_url = start_url
        else:
            # TODO: implement your next page logic to return None if no more pages
            raise NotImplementedError("Next page logic not implemented in base class")
        logger.info("Page url: %s", page_url)
        return page_url

    def _parse_page_links(
        self,
        page_url: str,
        print_every: int = 10,
        verbose: bool = False,
    ) -> Optional[List[dict]]:
        """Get the links from the given page."""

        # TODO: Parse the page and extract all links
        raise NotImplementedError("Parsing links is not implemented in base class")

    def _parse_article_text(self, url: str) -> dict:
        # TODO: Scrape the article page and extract the text
        raise NotImplementedError(
            "Parsing article text is not implemented in base class"
        )


def crawl_links(
    start_url: str,
    parse_page_func: Callable,
    next_page_func: Callable,
    start_page: int = 1,
    max_num_pages: Optional[int] = 2,
    link_urls: Optional[List[str]] = None,
    link_filepath: Optional[str] = None,
    delay_between_requests: float = 0.0,
) -> List[dict]:
    """Crawl links for article links with the given keyword.

    Args:
        keyword (str): Keyword to search for.
        search_url (str, optional): URL to search for the keyword. Defaults to "https://www.khmertimeskh.com/page/{page}/?s={keyword}".
        links (List[dict], optional): List of links to append to. Defaults to None.
        max_num_pages (Optional[int], optional): Maximum number of pages to crawl. Defaults to 2.
        link_filepath (Optional[str], optional): Filepath to save the links to. Defaults to None.
        print_every (int, optional): Print progress every n pages. Defaults to 10.
        verbose (bool, optional): Print progress. Defaults to False.

    Returns:
        List[dict]: List of links.
    """

    page = start_page
    page_cnt = 0
    page_url = None
    links = []
    link_urls = link_urls or []
    logger.info("Fetching links for url: %s", start_url)
    while True:
        # get next page url
        page_url = next_page_func(start_url, page_url, page)
        # Parse page
        page_links = parse_page_func(page_url)

        # Check if page_links is None
        if page_links is None:
            logger.info("No more links found, stopping...")
            break

        for link in page_links:
            if link["url"] not in link_urls:
                link["page_url"] = page_url
                link["page"] = page
                links.append(link)
                link_urls.append(link["url"])
                if link_filepath:
                    HyFI.append_to_jsonl(link, link_filepath)
            else:
                logger.info(
                    "Link %s already exists, skipping...",
                    link["url"],
                )

        page += 1
        page_cnt += 1

        if max_num_pages and page_cnt > max_num_pages:
            logger.info("Reached max number of pages, stopping...")
            break
        # Delay between requests
        if delay_between_requests > 0:
            logger.info("Sleeping for %s seconds...", delay_between_requests)
            time.sleep(delay_between_requests)

    logger.info("Finished fetching links for url: %s", start_url)
    logger.info("Total links fetched: %s", len(links))
    return links


def scrape_article_text(
    links: List[dict],
    parse_article_func: Callable,
    article_urls: Optional[List[dict]] = None,
    overwrite_existing: bool = False,
    max_num_articles: Optional[int] = 10,
    article_filepath: Optional[str] = None,
    delay_between_requests: float = 0.0,
    print_every: int = 10,
    verbose: bool = False,
) -> List[dict]:
    """Scrape the article text from the given links.

    Args:
        links (List[dict]): List of links to scrape.
        articles (Optional[List[dict]], optional): List of articles to append to. Defaults to None.
        overwrite_existing (bool, optional): Overwrite existing articles. Defaults to False.
        max_num_articles (Optional[int], optional): Maximum number of articles to scrape. Defaults to 10.
        article_filepath (Optional[str], optional): Filepath to save the articles to. Defaults to None.
        print_every (int, optional): Print progress every n articles. Defaults to 10.
        verbose (bool, optional): Print progress. Defaults to False.

    Returns:
        List[dict]: List of articles.
    """
    articles = []
    article_urls = article_urls or []
    for i, link in enumerate(links):
        if max_num_articles is not None and i >= max_num_articles:
            logger.info("Reached max number of articles, stopping...")
            break

        url = link["url"]
        title = link["title"]
        if url in article_urls and not overwrite_existing:
            logger.info("Article [%s](%s) already exists, skipping...", title, url)
            continue

        # Parse article
        _article = parse_article_func(url)
        if _article is None:
            logger.info(
                "Article [%s](%s) does not exist, skipping...",
                title,
                url,
            )
            continue
        article = link.copy()
        article.update(_article)
        articles.append(article)
        article_urls.append(url)
        if article_filepath:
            HyFI.append_to_jsonl(article, article_filepath)
        if (verbose and (i + 1) % print_every == 0) or delay_between_requests > 0:
            logger.info("Article [%s](%s) scraped", title, url)

        # Delay between requests
        if delay_between_requests > 0 and i < len(links) - 1:
            logger.info("Sleeping for %s seconds...", delay_between_requests)
            time.sleep(delay_between_requests)

    logger.info("Finished scraping articles")
    logger.info("Total articles scraped: %s", len(articles))
    return articles
